from stalimet.corpus import ParallelCorpus, SentencePair


class Scorer(object):

    def __init__(self, corpus: ParallelCorpus):
        self._corpus = corpus

    @staticmethod
    def __count_chunks(sentence_pair: SentencePair) -> int:
        previous_index = -1
        chunks = 1
        for a in sorted(sentence_pair.alignment):
            if previous_index != -1 and (a[1] is None or previous_index != a[1] - 1):
                chunks += 1
            previous_index = a[1]

        return chunks

    @staticmethod
    def __count_matching_chars(word1: str, word2: str) -> int:
        count = 0
        for i, c in enumerate(word1):
            if len(word2) > i and c == word2[i]:
                count += 1

        return count

    @staticmethod
    def __calculate_char_match(sentence_pair: SentencePair) -> float:
        matches = 0.0
        for a in sorted(sentence_pair.alignment):
            count = 0 if a[1] is None else Scorer.__count_matching_chars(sentence_pair.words[a[0]], sentence_pair.mots[a[1]])
            if count > 0:
                matches += count / len(sentence_pair.words[a[0]])
        return matches

    def calculate_scores(self):
        """
        calculates scores for every sentence of the corpus using the formula:

        score = (sum(matching_character_for_every_word) / number_of_words) * fragmentation_penalty
        fragmentation_penalty = number_of_continuous_chunks / (2 * sum(matching_character_for_every_word))

        and stores the score in the sentence pair class
        """
        for i, sentence_pair in enumerate(self._corpus):
            try:
                chunks = Scorer.__count_chunks(sentence_pair)
                match = Scorer.__calculate_char_match(sentence_pair)

                recall = match / len(sentence_pair.mots)
                penalty = 0 if chunks == 1 else 0.5 * float(chunks) / match

                score = recall * (1.0 - penalty)
            except ZeroDivisionError:
                score = 0

            sentence_pair.score = score



from nltk import AlignedSent

from stalimet.ibm2_exact import IBMModel2Exact


class SentencePair(AlignedSent):
    @property
    def score(self) -> float:
        return self._score

    @score.setter
    def score(self, value: float):
        self._score = value

    def __init__(self, words, mots, alignment=None):
        super(SentencePair, self).__init__(words, mots, alignment=alignment)
        self._score = 0.0

    def __str__(self):
        return str(self.score) + ' ||| ' + ' '.join(self.words) + ' ||| ' + ' '.join(self.mots)

    def align(self, model: IBMModel2Exact):
        model.align(self)


class ParallelCorpus(list):
    @property
    def alignment_model(self) -> IBMModel2Exact:
        return self._alignment_model

    @alignment_model.setter
    def alignment_model(self, model: IBMModel2Exact):
        self._alignment_model = model

    def __init__(self):
        super(ParallelCorpus, self).__init__()
        self._alignment_model = None
        self._self_trained  = False

    def build_parallel_corpus(self, tgt: list, ref: list) -> 'ParallelCorpus':
        for i in range(len(tgt)):
            self.append(SentencePair(tgt[i].split(), ref[i].split()))
        return self

    def train_alignment_model(self):
        self._alignment_model = IBMModel2Exact(self, 5)
        return self

    def align_sentences(self):
        if not self._self_trained:
            for sentence_page in self:
                self.alignment_model.align(sentence_page)




from nltk.tokenize import word_tokenize
import string


def run_preprocess(path, language, keep_punctuation=False):
    preprocess = Preprocess(open(path).readlines(), language).lowercase().tokenize()
    if not keep_punctuation:
        preprocess = preprocess.remove_punctuation()
    return preprocess.lines


class Preprocess:

    def __init__(self, lines: list, language: str):
        self.lines = [line.strip() for line in lines]
        self.preprocessor_run = []
        self.language = language

    def lowercase(self) -> 'Preprocess':
        for i, line in enumerate(self.lines):
            self.lines[i] = self.lines[i].lower()
        return self

    def remove_punctuation(self) -> 'Preprocess':
        translate_table = dict((ord(char), None) for char in string.punctuation)
        for i, line in enumerate(self.lines):
            self.lines[i] = self.lines[i].translate(translate_table)
        return self

    def tokenize(self) -> 'Preprocess':
        for i, line in enumerate(self.lines):
            self.lines[i] = ' '.join(word_tokenize(self.lines[i], self.language))
        return self

    def split(self) -> 'Preprocess':
        for i, line in enumerate(self.lines):
            self.lines[i] = self.lines[i].split()
        return self



from stalimet.corpus import ParallelCorpus
from stalimet.preprocess import run_preprocess
from stalimet.score import Scorer


class Stalimet:
    """
    Main metric class encapsulating training, aligning and scoring the parallel corpora of translation / reference
    segments, if extrenal train corpus is proved the alignment model is trained using this corpus,
    if not it is trained on the corpus for scoring
    """

    @property
    def train_corpus(self) -> ParallelCorpus:
        return self._train_corpus

    @property
    def corpus(self) -> ParallelCorpus:
        return self._corpus

    @property
    def language(self) -> str:
        return self._language

    def __init__(self, tgt_path: str, ref_path: str,
                 tgt_train_path: str = None, ref_train_path: str = None, language: str = 'english'):
        self._language = language
        self._corpus = ParallelCorpus()
        self._tgt_path = tgt_path
        self._ref_path = ref_path

        if not (tgt_train_path is None or ref_train_path is None):
            self._train_corpus = ParallelCorpus()
            self._tgt_train_path = tgt_train_path
            self._ref_train_path = ref_train_path
        else:
            self._train_corpus = None

    def train(self, keep_punctuation=False):
        tgt = run_preprocess(self._tgt_path, self._language, keep_punctuation)
        ref = run_preprocess(self._ref_path, self._language, keep_punctuation)
        self._corpus.build_parallel_corpus(tgt, ref)

        # if train corpus is specified we train alignment model on it and pass to main corpus,
        # if not train it on main corpus
        if self._train_corpus is not None:
            tgt_train = run_preprocess(self._tgt_train_path, self._language)
            ref_train = run_preprocess(self._ref_train_path, self._language)
            self._train_corpus.build_parallel_corpus(tgt_train, ref_train)
            self._train_corpus.train_alignment_model()
            self._corpus.alignment_model = self._train_corpus.alignment_model
        else:
            self._corpus.train_alignment_model()

        self._corpus.align_sentences()

    def score(self) -> list:
        scorer = Scorer(self._corpus)
        scorer.calculate_scores()

        scores = []
        for entry in self._corpus:
            scores.append(entry.score)

        return scores



from nltk.translate.ibm2 import IBMModel2, IBMModel
from nltk.translate import Alignment


class IBMModel2Exact(IBMModel2):
    """
    Modifies imb1 lexical translation model by giving maximum
    probability to exact matching words
    """

    def __init__(self, sentence_aligned_corpus, iterations, probability_tables=None):
        super(IBMModel2, self).__init__(sentence_aligned_corpus)

        if probability_tables is None:
            self.set_uniform_probabilities(sentence_aligned_corpus)
        else:
            # Set user-defined probabilities
            self.translation_table = probability_tables['translation_table']

        for n in range(0, iterations):
            self.train(sentence_aligned_corpus)

        self.__align_all(sentence_aligned_corpus)

    def __align_all(self, parallel_corpus):
        for sentence_pair in parallel_corpus:
            self.__align(sentence_pair)

    def __align(self, sentence_pair):
        """
        Determines the best word alignment for one sentence pair from
        the corpus that the model was trained on.

        The best alignment will be set in ``sentence_pair`` when the
        method returns. In contrast with the internal implementation of
        IBM models, the word indices in the ``Alignment`` are zero-
        indexed, not one-indexed.

        :param sentence_pair: A sentence in the source language and its
            counterpart sentence in the target language
        :type sentence_pair: AlignedSent
        """
        best_alignment = []

        for j, trg_word in enumerate(sentence_pair.words):
            # Initialize trg_word to align with the NULL token
            best_prob = max(self.translation_table[trg_word][None],
                            IBMModel.MIN_PROB)
            best_alignment_point = None
            for i, src_word in enumerate(sentence_pair.mots):
                align_prob = self.translation_table[trg_word][src_word]
                if trg_word == src_word:
                    align_prob = 1
                if align_prob >= best_prob:  # prefer newer word in case of tie
                    best_prob = align_prob
                    best_alignment_point = i

            best_alignment.append((j, best_alignment_point))

        sentence_pair.alignment = Alignment(best_alignment)

    def align(self, sentence_pair):
        self.__align(sentence_pair)


from setuptools import setup

from stalimet import __version__, __author__

setup(
   name="stalimet",
   version=__version__,
   packages=['stalimet'],
   install_requires=['nltk'],
   scripts=['stalimet_run'],
   author=__author__,
   author_email="mari.fomicheva@gmail.com, amalinovskiy@gmail.com",
   description="Machine translation quality metric based on IBM2 statistical alignment model",
   license="Apache License 2.0",
   keywords="machine translation quality metric",
   url="https://github.com/amalinovskiy/stalimet",
)